import os
from os import path

from .structures import Structures


def write_match_output(structures: Structures, odir_match: str) -> None:
    """
    Save Tregex's match output

    :param structures: an instance of Structures
    :param dir_match_output: where to save the match output
    """
    bn_input = path.basename(structures.ifile)
    bn_input_rstripped = path.splitext(bn_input)[0]
    # bn_input with the extension at the right side stripped
    subdir_match_output = path.join(odir_match, bn_input_rstripped)
    if not path.isdir(subdir_match_output):
        # if not exists and is a directory
        os.makedirs(subdir_match_output)
    for structure in structures.to_search_for:
        bn_match_output = bn_input_rstripped + "-" + structure.name + ".matches"
        fn_match_output = path.join(subdir_match_output, bn_match_output)
        if structure.matches:
            with open(fn_match_output, "w", encoding="utf-8") as f:
                f.write(structure.matches)
    print(f"Match output was saved to {subdir_match_output}")


def write_freq_output(freq_output: str, ofile_freq: str) -> None:
    """
    :param freq_output: comma-separated frequency output
    :param fn_output: where to save the frequency output
    """
    with open(ofile_freq, "w", encoding="utf-8") as f:
        f.write(f"{Structures.fields}\n{freq_output}")
    print(f"Frequency output was saved to {ofile_freq}. Done.")



import os
from os import path
import re
import subprocess
import sys
from typing import Tuple, Generator

from .structures import Structures


class Analyzer:
    method_parser = "edu.stanford.nlp.parser.lexparser.LexicalizedParser"
    model_parser = "edu/stanford/nlp/models/lexparser/englishPCFG.ser.gz"
    method_tregex = "edu.stanford.nlp.trees.tregex.TregexPattern"

    def __init__(self, dir_stanford_parser, dir_stanford_tregex):
        """
        :param dir_parser: directory to Stanford Parser
        :param dir_tregex: directory to Tregex
        """
        self.classpath_parser = '"' + dir_stanford_parser + os.sep + "*" + '"'
        self.classpath_tregex = (
            '"' + dir_stanford_tregex + os.sep + "stanford-tregex.jar" + '"'
        )

    def _parse(self, ifile: str, fn_parsed: str) -> None:
        """
        Call Stanford Parser

        :param ifile: file to parse
        :param fn_parsed: where to save the parsed results
        :return: None
        """
        cmd = (
            f"java -mx1500m -cp {self.classpath_parser} {self.method_parser} "
            f"-outputFormat penn {self.model_parser} {ifile} > {fn_parsed}"
        )
        if not path.exists(fn_parsed):
            # print(f"{fn_parsed} does not exist, running Stanford Parser...")
            subprocess.run(cmd, shell=True, capture_output=True)
            return
        mt_input = path.getmtime(ifile)  # get the last modification time
        mt_parsed = path.getmtime(fn_parsed)
        if mt_input > mt_parsed:
            # print(
            #     f"{fn_parsed} is older than {ifile}, "
            #     + "running Stanford Parser..."
            # )
            subprocess.run(cmd, shell=True, capture_output=True)
        # else:
        #     print(f"{fn_parsed} is newer than {ifile}, parsing is skipped.")

    def _query(self, pattern: str, fn_parsed: str) -> Tuple[int, str]:
        """
        Call Tregex to query {pattern} against {fn_parsed}

        :param pattern: Tregex pattern
        :param fn_parsed: parsed file by Stanford Parser
        :return (int) frequency: frequency of the pattern
        :return (str) matched_subtreees: matched subtrees of the pattern
        """
        cmd = (
            "java -mx100m -cp"
            f" {self.classpath_tregex} {self.method_tregex} {pattern} {fn_parsed} -o"
        )
        p = subprocess.run(cmd, shell=True, capture_output=True)
        match_reslt = re.search(
            r"There were (\d+) matches in total\.", p.stderr.decode()
        )
        if match_reslt:
            freq = match_reslt.group(1)
        else:
            os.remove(fn_parsed)
            # Remove fn_parsed to make sure parsing will not be skipped on next running.
            sys.exit(
                "Error: failed to obtain frequency. It is likely that:\n"
                "(1) Tregex's interface has"
                " changed. As a workaround, download an older version, v4.2.0,"
                " for example. If Tregex's interface does have changed, the"
                " latest version of Tregex will be supported in next few"
                " releases.\n"
                "(2) You manually modified Tregex patterns in"
                " structures.py. Make sure that: on Windows, those patterns are"
                " enclosed by double quotes; on Linux and macOS they are"
                " surrounded by single quotes."
            )
        matched_subtrees = p.stdout.decode()
        matched_subtrees = self._add_terms(matched_subtrees)
        return int(freq), matched_subtrees

    def _add_terms(self, subtrees: str) -> str:
        """
        Add terminals above each subtree

        :param subtrees: matched subtrees by Tregex
         e.g. (NP (NNP Mr.) (NNP Reed))
        :return: e.g.
         Mr. Reed
         (NP (NNP Mr.) (NNP Reed))
        """
        result = ""
        pattern = r"([^\s)]+)\)"
        for subtree in re.split(r"(?:\r?\n){2,}", subtrees):
            terminals = " ".join(re.findall(pattern, subtree))
            result += f"{terminals}\n{subtree}\n\n"
        return result.strip()

    def _analyze_text(self, ifile, reserve_parsed) -> Structures:
        """
        Analyze a text file

        :param ifile: which file to analyze
        :param reserve_parsed: option to reserve Stanford Parser's
         parsing results
        :return structures: an instance of Structures
        """
        fn_parsed = path.splitext(ifile)[0] + ".parsed"
        self._parse(ifile, fn_parsed)

        structures = Structures(path.basename(ifile))
        for structure in structures.to_search_for:
            print(
                f'\t[Tregex] Querying "{structure.desc}" against '
                f"{path.basename(fn_parsed)}..."
            )
            structure.freq, structure.matches = self._query(
                structure.pat, fn_parsed
            )
        structures.update_freqs()

        with open(fn_parsed) as f:
            structures.W.freq = len(
                re.findall(r"\([A-Z]+\$? [^()]+\)", f.read())
            )

        structures.compute_SC_indicies()
        if not reserve_parsed:
            os.remove(fn_parsed)
        return structures

    def perform_analysis(
        self, ifiles: list, reserve_parsed: bool
    ) -> Generator[Structures, None, None]:
        """
        :param ifiles: list of input files
        :param reserve_parsed: option to reserve Stanford Parser's
         parsing results
        """
        total = len(ifiles)
        for i, ifile in enumerate(ifiles):
            print(
                f"[NeoSCA] Processing {path.basename(ifile)} ({i+1}/{total})..."
            )
            structures = self._analyze_text(ifile, reserve_parsed)
            yield structures



import argparse
import glob
from os import path
import os
import subprocess
import sys
from typing import List, Optional, Tuple

from . import __version__
from .utils.analyzer import Analyzer
from .utils.writer import write_match_output
from .utils.writer import write_freq_output

# For all the procedures in SCAUI, return a tuple as the result
# The first element bool indicates whether the procedure succeeds
# The second element is the error message if it fails.
SCAProcedureResult = Tuple[bool, Optional[str]]


class SCAUI:
    def __init__(self):
        self.parser: argparse.ArgumentParser = self.create_parser()

        self.ofile_freq: str = "result.csv"
        self.dir_stanford_parser: Optional[str] = os.getenv(
            "STANFORD_PARSER_HOME"
        )
        self.dir_stanford_tregex: Optional[str] = os.getenv(
            "STANFORD_TREGEX_HOME"
        )
        self.reserve_parsed: bool = False
        self.reserve_match: bool = False

        self.options: argparse.Namespace = argparse.Namespace()

    def create_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            prog="nsca",
            formatter_class=lambda prog: argparse.HelpFormatter(
                prog, max_help_position=50, width=100
            ),
        )
        parser.add_argument(
            "--version",
            action="store_true",
            default=False,
            help="show version of NeoSCA",
        )
        parser.add_argument(
            "-o",
            "--output",
            default=None,
            help="output file",
        )
        parser.add_argument(
            "--parser",
            dest="dir_stanford_parser",
            default=None,
            help=(
                "directory to Stanford Parser, defaults to STANFORD_PARSER_HOME"
            ),
        )
        parser.add_argument(
            "--tregex",
            dest="dir_stanford_tregex",
            default=None,
            help=(
                "directory to Stanford Tregex, defaults to STANFORD_TREGEX_HOME"
            ),
        )
        parser.add_argument(
            "-p",
            "--reserve-parsed",
            dest="reserve_parsed",
            action="store_true",
            default=False,
            help="option to reserve parsed files by Stanford Parser",
        )
        parser.add_argument(
            "-m",
            "--reserve-match",
            dest="reserve_match",
            default=False,
            action="store_true",
            help="option to reserve match results by Stanford Tregex",
        )
        return parser

    def parse(self, argv: List[str]) -> SCAProcedureResult:
        args = argv[1:] if argv[1:] else ["--help"]
        options, ifile_list = self.parser.parse_known_args(args)

        if options.output:
            self.ofile_freq = options.output
        self.odir_match = path.splitext(self.ofile_freq)[0]

        if options.dir_stanford_parser:
            self.dir_stanford_parser = options.dir_stanford_parser
        if self.dir_stanford_parser is None or not path.isdir(
            self.dir_stanford_parser
        ):
            return False, "Stanford Parser not found."

        if options.dir_stanford_tregex:
            self.dir_stanford_tregex = options.dir_stanford_tregex
        if self.dir_stanford_tregex is None or not path.isdir(
            self.dir_stanford_tregex
        ):
            return False, "Stanford Tregex not found."

        if options.reserve_parsed:
            self.reserve_parsed = options.reserve_parsed
        if options.reserve_match:
            self.reserve_match = options.reserve_match

        self.options, self.ifile_list = options, ifile_list

        valid_ifile_list = []
        for f in self.ifile_list:
            if path.isfile(f):
                valid_ifile_list.append(f)
            elif glob.glob(f):
                valid_ifile_list.extend(glob.glob(f))
            else:
                return (
                    False,
                    "The following file either does not exist or is not a"
                    f" regular file: \n\n{f}",
                )
        self.ifile_list = valid_ifile_list

        return True, None

    def _has_java(self) -> bool:
        try:
            subprocess.run(
                "java --version", shell=True, check=True, capture_output=True
            )
        except subprocess.CalledProcessError:
            return False
        return True

    def run_analyzer(self) -> SCAProcedureResult:
        if not self._has_java():
            return (
                False,
                "Error: Java is unavailable.\n\n1. To install it, visit"
                " https://www.java.com/en/download.\n2. After installing, make"
                " sure you can access it in the cmd window by typing in `java"
                " --version`.",
            )
        if not self.ifile_list:
            return False, "Input files are not provided."
        analyzer = Analyzer(self.dir_stanford_parser, self.dir_stanford_tregex)
        structures_list = list(
            analyzer.perform_analysis(self.ifile_list, self.reserve_parsed)
        )  # list of instances of Structures, each for one corresponding input file

        freq_output = ""
        for structures in structures_list:
            freq_output += structures.get_freqs() + "\n"
        write_freq_output(freq_output, self.ofile_freq)

        if self.reserve_match:
            for structures in structures_list:
                write_match_output(structures, self.odir_match)

        return True, None

    def run(self) -> SCAProcedureResult:
        if self.options.version:
            return self.show_version()
        else:
            return self.run_analyzer()

    def show_version(self) -> SCAProcedureResult:
        print(__version__)
        return True, None


def main():
    ui = SCAUI()
    success, err_msg = ui.parse(sys.argv)
    if not success:
        print(err_msg)
        sys.exit(1)
    success, err_msg = ui.run()
    if not success:
        print(err_msg)
        sys.exit(1)
